TASK_SUBMITTED: str = "morpfw.task_submitted"
TASK_STARTING: str = "morpfw.task_starting"
TASK_COMPLETED: str = "morpfw.task_completed"
TASK_FAILED: str = "morpfw.task_failed"
TASK_FINALIZED: str = "morpfw.task_finalized"
SCHEDULEDTASK_STARTING: str = "morpfw.scheduler.task_starting"
SCHEDULEDTASK_COMPLETED: str = "morpfw.scheduler.task_completed"
SCHEDULEDTASK_FAILED: str = "morpfw.scheduler.task_failed"
SCHEDULEDTASK_FINALIZED: str = "morpfw.scheduler.task_finalized"
